@@ -1,6 +1,7 @@
 import requests
 import time
 import pandas
+from string import Template
 
 
 ENDPOINT = 'https://api.portfolio123.com'
@@ -14,6 +15,7 @@ DATA_PATH = '/data'
 RANK_RANKS_PATH = '/rank/ranks'
 RANK_PERF_PATH = '/rank/performance'
 DATA_UNIVERSE_PATH = '/data/universe'
+STRATEGY_UNIVERSE_PATH = Template('/strategy/$id')
 
 
 class ClientException(Exception):
@@ -100,7 +102,7 @@ class Client(object):
                 message = ': ' + message
             raise ClientException(f'API authentication failed{message}', resp=resp)
 
-    def _req_with_auth_fallback(self, *, name: str, url: str, params, stop: bool = False):
+    def _req_with_auth_fallback(self, *, name: str, method: str = 'POST', url: str, params=None, stop: bool = False):
         """
         Request with authentication fallback, used by all requests (except authentication)
         :param name: request action
@@ -111,18 +113,27 @@ class Client(object):
         """
         resp = None
         if self._session.headers.get('Authorization') is not None:
-            resp = req_with_retry(
-                self._session.post,
-                self._max_req_retries,
-                url=url,
-                json=params,
-                verify=self._verify_requests,
-                timeout=self._timeout
-            )
+            if method == 'POST':
+                resp = req_with_retry(
+                    self._session.post,
+                    self._max_req_retries,
+                    url=url,
+                    json=params,
+                    verify=self._verify_requests,
+                    timeout=self._timeout
+                )
+            else:
+                resp = req_with_retry(
+                    self._session.get,
+                    self._max_req_retries,
+                    url=url,
+                    verify=self._verify_requests,
+                    timeout=self._timeout
+                )
         if resp is None or resp.status_code == 403:
             if not stop:
                 self.auth()
-                return self._req_with_auth_fallback(name=name, url=url, params=params, stop=True)
+                return self._req_with_auth_fallback(name=name, method=method, url=url, params=params, stop=True)
         elif resp.status_code == 200:
             return resp
         else:
@@ -175,11 +186,11 @@ class Client(object):
         :param params:
         :return:
         """
-        self._req_with_auth_fallback(
+        return self._req_with_auth_fallback(
             name='universe update',
             url=self._endpoint + UNIVERSE_PATH,
             params=params
-        )
+        ).json()
 
     def rank_update(self, params: dict):
         """
@@ -187,11 +198,11 @@ class Client(object):
         :param params:
         :return:
         """
-        self._req_with_auth_fallback(
+        return self._req_with_auth_fallback(
             name='ranking system update',
             url=self._endpoint + RANK_PATH,
             params=params
-        )
+        ).json()
 
     def data(self, params: dict, to_pandas: bool = False):
         """
@@ -212,25 +223,22 @@ class Client(object):
             with_name = params.get('includeNames')
             data = []
             for date_idx, date in enumerate(ret['dates']):
-                row = []
                 for item_uid, item_data in ret['items'].items():
-                    row.append(date)
-                    row.append(item_uid)
-                    row.append(item_data['ticker'])
+                    row = [date, item_uid, item_data['ticker']]
                     if with_cusips:
                         row.append(item_data['cusip'])
                     if with_name:
                         row.append(item_data['name'])
                     for formula_idx, formula in enumerate(params['formulas']):
                         row.append(item_data['series'][formula_idx][date_idx])
-                data.append(row)
+                    data.append(row)
             columns = ['date', 'p123Uid', 'ticker']
             if with_cusips:
                 columns.append('cusip')
             if with_name:
                 columns.append('name')
-            for formula in params['formulas']:
-                columns.append(formula)
+            for formula_idx, formula in enumerate(params['formulas']):
+                columns.append(f'formula{formula_idx + 1}')
             ret = pandas.DataFrame(data=data, columns=columns)
             ret.attrs['raw_obj'] = raw_obj
 
@@ -244,7 +252,7 @@ class Client(object):
         :return:
         """
         ret = self._req_with_auth_fallback(
-            name='data',
+            name='data universe',
             url=self._endpoint + DATA_UNIVERSE_PATH,
             params=params
         ).json()
@@ -252,8 +260,8 @@ class Client(object):
         if to_pandas:
             raw_obj = dict(ret)
             for formula_idx, formula in enumerate(params['formulas']):
-                ret[formula] = ret['data'][formula_idx]
-            del ret['quota'], ret['quotaRemaining'], ret['data']
+                ret[f'formula{formula_idx + 1}'] = ret['data'][formula_idx]
+            del ret['cost'], ret['quotaRemaining'], ret['data']
             if ret.get('dt'):
                 del ret['dt']
             ret = pandas.DataFrame(data=ret)
@@ -269,14 +277,14 @@ class Client(object):
         :return:
         """
         ret = self._req_with_auth_fallback(
-            name='data',
+            name='ranking system ranks',
             url=self._endpoint + RANK_RANKS_PATH,
             params=params
         ).json()
 
         if to_pandas:
             raw_obj = dict(ret)
-            del ret['quota'], ret['quotaRemaining'], ret['dt']
+            del ret['cost'], ret['quotaRemaining'], ret['dt']
             nodes = ret.get('nodes')
             if nodes is not None:
                 for node_idx, node_name in enumerate(nodes['names']):
@@ -289,6 +297,7 @@ class Client(object):
             additional_data = ret.get('additionalData')
             if additional_data is not None:
                 for data_idx, data_name in enumerate(params['additionalData']):
+                    data_name = f'formula{data_idx + 1}'
                     ret[data_name] = []
                     for idx, uid in enumerate(ret['p123Uids']):
                         ret[data_name].append(additional_data[idx][data_idx])
@@ -305,11 +314,23 @@ class Client(object):
         :return:
         """
         return self._req_with_auth_fallback(
-            name='data',
+            name='ranking system performance',
             url=self._endpoint + RANK_PERF_PATH,
             params=params
         ).json()
 
+    def strategy(self, strategy_id: int):
+        """
+        Strategy details
+        :param strategy_id:
+        :return:
+        """
+        return self._req_with_auth_fallback(
+            name='strategy details',
+            method='GET',
+            url=self._endpoint + STRATEGY_UNIVERSE_PATH.substitute(id=strategy_id)
+        ).json()
+
     def get_api_id(self):
         return self._api_id
 
